#ifndef ITC_H
#define ITC_H

#include <iostream>
#include <string>
#include <queue>
#include <cassert>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

namespace itc {

typedef boost::function<void (void)> CallType;

struct ProcessingPolicy
{
	virtual void queue(CallType call) = 0;
	virtual void execute() = 0;
	virtual ~ProcessingPolicy() {}
};

class WaitAndExecute : public ProcessingPolicy
{
public:
	WaitAndExecute():
		m_mutex(), m_cond(), m_req(false), m_callQ()
	{}	

	void queue(CallType call)
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		m_callQ.push(call);
		m_req = true;
		m_cond.notify_one();
	}
	
	void execute()
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		
		// wait
		while (!m_req)
		{
			m_cond.wait(lock);
		}
		
		// execute
		while (!m_callQ.empty())
		{
			m_callQ.front()();
			m_callQ.pop();
		}
		m_req = false;	
	}	

private:
	boost::mutex m_mutex;
	boost::condition_variable m_cond;
	bool m_req;
	std::queue<CallType>  m_callQ;	
};

class ProcessAndExecute : public ProcessingPolicy
{
public:
	ProcessAndExecute() : m_mutex(), m_callQ()	{}
	
	void queue(CallType call)
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		m_callQ.push(call);
	}
	
	void execute()
	{
		CallType call;
		bool req = false;
		
		// process
		process();
		
		// execute
		{
			boost::unique_lock<boost::mutex> lock(m_mutex);
			if (!m_callQ.empty())
			{
				call = m_callQ.front();
				m_callQ.pop();
				req = true;
			}
		}
		
		if (req)
			call();		
	}	

private:
	virtual void process() = 0;

	boost::mutex m_mutex;
	std::queue<CallType>  m_callQ;
};

class ITCThreadBase
{
public:
	ITCThreadBase(boost::shared_ptr<ProcessingPolicy> policy, const std::string name) :
		m_thr(), m_policy(policy),
		m_name(name), m_run(true), m_code(0)
	{
		m_thr = boost::thread(&ITCThreadBase::run, this);
	}

	const std::string& getName()
	{
		return m_name;
	}

	void quit(int code = 0)
	{
		queue(boost::bind(&ITCThreadBase::doQuit, this, code));
	}

	int join()
	{
		m_thr.join();
		return m_code;
	}

	void queue(CallType call)
	{
		m_policy->queue(call);
	}

private:
	void run()
	{
		while (m_run)
		{
			m_policy->execute();
		}
	}

	void doQuit(int code)
	{
		m_run = false;
		m_code = code;
	}

	boost::thread m_thr;
	boost::shared_ptr<ProcessingPolicy> m_policy;
	const std::string m_name;
	bool m_run;
	int m_code;
};

template <typename DerivedThread, typename PolicyType = WaitAndExecute>
struct ITCThread : public ITCThreadBase
{
	ITCThread(const std::string& name = "itcthr") :
		ITCThreadBase(boost::shared_ptr<ProcessingPolicy>(new PolicyType()), name)
	{}

	void doIt(void (DerivedThread::*mf)(void))
	{
		queue(boost::bind(mf, (DerivedThread*)this));
	}

	template <typename ArgType>
	void doIt(void (DerivedThread::*mf)(ArgType), ArgType arg)
	{
		queue(boost::bind(mf, (DerivedThread*)this, arg));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (DerivedThread::*mf)(ArgType1, ArgType2),
			  ArgType1 arg1, ArgType2 arg2)
	{
		queue(boost::bind(mf, (DerivedThread*)this, arg1, arg2));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (DerivedThread::*mf)(ArgType1, ArgType2, ArgType3),
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		queue(boost::bind(mf, (DerivedThread*)this, arg1, arg2, arg3));
	}
};

template <typename Callee, typename PolicyType = WaitAndExecute>
class ITCAdaptor : public ITCThreadBase
{
public:
	ITCAdaptor(boost::shared_ptr<Callee> callee, const std::string& name = "itcapt") :
		ITCThreadBase(boost::shared_ptr<ProcessingPolicy>(new PolicyType()), name), m_callee(callee)
	{}

	void doIt(void (Callee::*mf)(void))
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee));
	}

	template <typename ArgType>
	void doIt(void (Callee::*mf)(ArgType), ArgType arg)
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee, arg));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2), 
			  ArgType1 arg1, ArgType2 arg2)
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee, arg1, arg2));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2, ArgType3),
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee, arg1, arg2, arg3));
	}

private:
	boost::shared_ptr<Callee> m_callee;
};

template <typename PolicyType = WaitAndExecute>
struct FunctionITCAdaptor : public ITCThreadBase
{
	FunctionITCAdaptor(const std::string& name = "fnitcapt") :
		ITCThreadBase(boost::shared_ptr<ProcessingPolicy>(new PolicyType()), name)
	{}

    void doIt(void (*func)(void))
    {
        queue(func);
    }

    template <typename ArgType>
    void doIt(void (*func)(ArgType), ArgType arg)
    {
        queue(boost::bind(func, arg));
    }

    template <typename ArgType1, typename ArgType2>
    void doIt(void (*func)(ArgType1, ArgType2), ArgType1 arg1, ArgType2 arg2)
    {
        queue(boost::bind(func, arg1, arg2));
    }

    template <typename ArgType1, typename ArgType2, typename ArgType3>
    void doIt(void (*func)(ArgType1, ArgType2, ArgType3),
              ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
    {
        queue(boost::bind(func, arg1, arg2, arg3));
    }
};

};

#endif
