/// @file main.cpp

#include <iostream>
#include <string>
#include <queue>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include "itc.h"

using namespace std;
using namespace boost;
using namespace boost::posix_time;
using namespace itc;

// example #1 : new class which inherits from ITCThread 
class HelloThread : public ITCThread<HelloThread>
{
public:
	HelloThread(const string& name = "hello") :
		ITCThread<HelloThread>(name) {}
	
	void hello(const string& name)
	{
		doIt(&HelloThread::doHello, name);
	}

	void yahoo()
	{
		doIt(&HelloThread::doYahoo);
	}

private:
	void doHello(const string name)
	{
		cout << getName() << ": Hello~ " << name << endl;
	}
	
	void doYahoo()
	{
		cout << getName() << ": Yahoo~~~~~!" << endl;
	}
};

// example #2 : new class which inherits from ITCThread 
class PrintVector : public ITCThread<PrintVector>
{
public:
	PrintVector(const string& name = "prvec") :
		ITCThread<PrintVector>(name)
		{}
	
	void print(const vector<int>& vi)
	{
		doIt(&PrintVector::doPrint, vi);
	}

	void printTwoVectors(const vector<int>& vi, const vector<double>& vd)
	{
		doIt(&PrintVector::doPrintTwoVectors, vi, vd);
	}

private:
	void doPrint(const vector<int> vi)
	{
		copy(vi.begin(), vi.end(), ostream_iterator<int>(cout, " "));
		cout << endl;
	}

	void doPrintTwoVectors(const vector<int> vi, const vector<double> vd)
	{
		copy(vi.begin(), vi.end(), ostream_iterator<int>(cout, " "));
		cout << endl;
		copy(vd.begin(), vd.end(), ostream_iterator<double>(cout, " "));
		cout << endl;
	}
};

// example #3: legacy class + wrapping member function
struct Hello
{
	void sayHelloTo(const string name)
	{
		cout << "Hello, " << name << "!" << endl;
	}
};

struct HelloITCAdaptor : public ITCAdaptor<Hello>
{
	HelloITCAdaptor(shared_ptr<Hello> hello, const string& name) : 
		ITCAdaptor<Hello>(hello, name) {}

	void sayHelloTo(const string& name)
	{
		doIt(&Hello::sayHelloTo, name);
	}
};

// example #4: legacy class, no wrapping member function
struct ButtonHandler
{
	void onClick(int x, int y)
	{
		cout << "button clicked at (" << x << ", " << y << ")" << endl;
	}
};

typedef ITCAdaptor<ButtonHandler> ButtonHandlerITCAdaptor;

// example #5
void printHello()
{
	cout << "Hello, world" << endl;
}

void sayHelloTo(const string name)
{
	cout << "Hello, " << name << "!" << endl;
}

void sayWelcomeToAndGoodByeTo(const string name1, const string name2)
{
	cout << "Welcome, " << name1 << "!" << endl;
	cout << "Good-bye, " << name2 << "!" << endl;
}

// example #6
struct Greeter : public FunctionITCAdaptor<>
{
	Greeter(const string& name) : FunctionITCAdaptor<>(name) {}

	void printHello()
	{
		doIt(::printHello);
	}

	void sayHelloTo(const string& name)
	{
		doIt(::sayHelloTo, name);
	}

	void sayWelcomeToAndGoodByeTo(const string& name1, const string& name2)
	{
		doIt(::sayWelcomeToAndGoodByeTo, name1, name2);
	}
};

// example #7
class ControllerInternalProcessor : public ProcessAndExecute
{
private:
	virtual void process()
	{
		static int cnt = 0;
		if (cnt++ % 100 == 0)
		{
			cout << "processing..." << endl;
		}
		this_thread::sleep(millisec(100));
	}
};

class Controller : public ITCThread<Controller, ControllerInternalProcessor>
{
public:
	Controller(const string& name) : 
		ITCThread<Controller, ControllerInternalProcessor>(name) 
	{}
	
	void set(const string& target, double value)
	{
		doIt(&Controller::doSet, target, value);
	}

private:
	void doSet(const string target, double value)
	{
		cout << getName() << ": setting " << target
			 << "'s value to " << value << endl;
	}
};

// example #8
struct Base
{
	virtual void print()
	{
		cout << "Base::print()" << endl;
	}
};

struct Derived : public Base
{
	virtual void print()
	{
		cout << "Derived::print()" << endl;
	}
};

struct BaseITCAdaptor : public ITCAdaptor<Base>
{
	BaseITCAdaptor(shared_ptr<Base> b, const string& name) :
		ITCAdaptor<Base>(b, name)
	{}

	void print()
	{
		doIt(&Base::print);
	}
};

int
main()
{
	// example #8
	BaseITCAdaptor bitcthr(shared_ptr<Base>(new Base()), "baseitcthr");
	BaseITCAdaptor ditcthr(shared_ptr<Derived>(new Derived()), "deriveditcthr");
	bitcthr.print();
	ditcthr.print();
	bitcthr.quit(), ditcthr.quit();
	bitcthr.join(), ditcthr.join();

#if 1
	// example #7
	string name = "aaa";
	double value = 1.1;
	Controller ctrl("CTRL");

	do
	{
		cout << "enter name and value: ";
		cin >> name >> value;
		cout << "setting " << name << "'s value to " << value << endl;
		ctrl.set(name, value);
		this_thread::sleep(millisec(100));
	} while(cin);

	for (int i = 0; i < 100; ++i)
	{
		ctrl.set(name, value);
		this_thread::sleep(millisec(10));
	}
	ctrl.quit();
	ctrl.join();
#endif

#if 1
	// example #1
	HelloThread itc1("H1");
	HelloThread itc2("H2");
	
	for (int i = 0; i < 10; ++i)
	{
		this_thread::sleep(millisec(1000));
		itc1.hello("world!"), itc1.yahoo();
		this_thread::sleep(millisec(50));
		itc2.hello("body!"), itc2.yahoo();
	}
	this_thread::sleep(millisec(1000));
	itc1.quit(1);
	this_thread::sleep(millisec(50));
	itc2.quit(2);
	itc1.join(), itc2.join();

	// example #2
	string s = "prvec";
	PrintVector prvec(s);

	vector<int> vi;
	for (int i = 0; i < 10; ++i)
	{
		vi.push_back(i);
	}
	prvec.print(vi);

	vector<double> vd;
	for (int i = 0; i < 10; ++i)
	{
		vd.push_back(i * 10.5);
	}
	prvec.printTwoVectors(vi, vd);

	prvec.quit();
	prvec.join();
	
	// example #3
	s = "hello";
	HelloITCAdaptor hello(shared_ptr<Hello>(new Hello()), s);
	for (int i = 0; i < 1000; ++i)
	{
		hello.sayHelloTo("Yoonsoo Kim");
	}
	hello.quit();
	hello.join();

	// example #4
	s = "btn_handler";
	ButtonHandlerITCAdaptor btnHdlr(shared_ptr<ButtonHandler>(new ButtonHandler()), s);
	btnHdlr.doIt(&ButtonHandler::onClick, 10, 20);
	btnHdlr.quit();
	btnHdlr.join();
	
	// example #5
	s = "fthr";
	FunctionITCAdaptor<> fthr(s);
	fthr.doIt(printHello);
	fthr.doIt(sayHelloTo, string("Yoonsoo Kim"));
	fthr.doIt(sayWelcomeToAndGoodByeTo, string("AAA"), string("BBB"));
	fthr.quit();
	fthr.join();

	// example #6
	s = "GRTR";
	Greeter g(s);
	g.printHello();
	g.sayHelloTo("YYY");
	g.sayWelcomeToAndGoodByeTo("YYY", "KKK");
	g.quit(100);
	int ret = g.join();
	cout << g.getName() << ": returned " << ret << endl; 
#endif

	return 0;
}

